/*This source code copyrighted by Lazy Foo' Productions 2004-2024
and may not be redistributed without written permission.*/

/* Headers */
//Using SDL, SDL_image, and STL string
#include <SDL3/SDL.h>
#include <SDL3/SDL_main.h>
#include <SDL3_image/SDL_image.h>
#include <string>


/* Constants */
//Screen dimension constants
constexpr int kScreenWidth{ 640 };
constexpr int kScreenHeight{ 480 };


/* Class Prototypes */
class LTexture
{
public:
    //Symbolic constant
    static constexpr float kOriginalSize = -1.f;

    //Initializes texture variables
    LTexture();

    //Cleans up texture variables
    ~LTexture();

    //Loads texture from disk
    bool loadFromFile( std::string path );

    //Cleans up texture
    void destroy();

    //Sets color modulation
    void setColor( Uint8 r, Uint8 g, Uint8 b);

    //Sets opacity
    void setAlpha( Uint8 alpha );

    //Sets blend mode
    void setBlending( SDL_BlendMode blendMode );

    //Draws texture
    void render( float x, float y, SDL_FRect* clip = nullptr, float width = kOriginalSize, float height = kOriginalSize, double degrees = 0.0, SDL_FPoint* center = nullptr, SDL_FlipMode flipMode = SDL_FLIP_NONE );

    //Gets texture dimensions
    int getWidth();
    int getHeight();

private:
    //Contains texture data
    SDL_Texture* mTexture;

    //Texture dimensions
    int mWidth;
    int mHeight;
};


/* Function Prototypes */
//Starts up SDL and creates window
bool init();

//Loads media
bool loadMedia();

//Frees media and shuts down SDL
void close();


/* Global Variables */
//The window we'll be rendering to
SDL_Window* gWindow{ nullptr };

//The renderer used to draw to the window
SDL_Renderer* gRenderer{ nullptr };

//The texture we're going to be modulating and blending
LTexture gColorsTexture; 

/* Class Implementations */
//LTexture Implementation
LTexture::LTexture():
    //Initialize texture variables
    mTexture{ nullptr },
    mWidth{ 0 },
    mHeight{ 0 }
{

}

LTexture::~LTexture()
{
    //Clean up texture
    destroy();
}

bool LTexture::loadFromFile( std::string path )
{
    //Clean up texture if it already exists
    destroy();

    //Load surface
    if( SDL_Surface* loadedSurface = IMG_Load( path.c_str() ); loadedSurface == nullptr )
    {
        SDL_Log( "Unable to load image %s! SDL_image error: %s\n", path.c_str(), SDL_GetError() );
    }
    else
    {
        //Color key image
        if( !SDL_SetSurfaceColorKey( loadedSurface, true, SDL_MapSurfaceRGB( loadedSurface, 0x00, 0xFF, 0xFF ) ) )
        {
            SDL_Log( "Unable to color key! SDL error: %s", SDL_GetError() );
        }
        else
        {
            //Create texture from surface
            if( mTexture = SDL_CreateTextureFromSurface( gRenderer, loadedSurface ); mTexture == nullptr )
            {
                SDL_Log( "Unable to create texture from loaded pixels! SDL error: %s\n", SDL_GetError() );
            }
            else
            {
                //Get image dimensions
                mWidth = loadedSurface->w;
                mHeight = loadedSurface->h;
            }
        }
        
        //Clean up loaded surface
        SDL_DestroySurface( loadedSurface );
    }

    //Return success if texture loaded
    return mTexture != nullptr;
}

void LTexture::destroy()
{
    //Clean up texture
    SDL_DestroyTexture( mTexture );
    mTexture = nullptr;
    mWidth = 0;
    mHeight = 0;
}

void LTexture::setColor( Uint8 r, Uint8 g, Uint8 b )
{
    SDL_SetTextureColorMod( mTexture, r, g, b );
}

void LTexture::setAlpha( Uint8 alpha )
{
    SDL_SetTextureAlphaMod( mTexture, alpha );
}

void LTexture::setBlending( SDL_BlendMode blendMode )
{
    SDL_SetTextureBlendMode( mTexture, blendMode );
}

void LTexture::render( float x, float y, SDL_FRect* clip, float width, float height, double degrees, SDL_FPoint* center, SDL_FlipMode flipMode )
{
    //Set texture position
    SDL_FRect dstRect = { x, y, static_cast<float>( mWidth ), static_cast<float>( mHeight ) };

    //Default to clip dimensions if clip is given
    if( clip != nullptr )
    {
        dstRect.w = clip->w;
        dstRect.h = clip->h;
    }

    //Resize if new dimensions are given
    if( width > 0 )
    {
        dstRect.w = width;
    }
    if( height > 0 )
    {
        dstRect.h = height;
    }

    //Render texture
    SDL_RenderTextureRotated( gRenderer, mTexture, clip, &dstRect, degrees, center, flipMode );
}

int LTexture::getWidth()
{
    return mWidth;
}

int LTexture::getHeight()
{
    return mHeight;
}


/* Function Implementations */
bool init()
{
    //Initialization flag
    bool success{ true };

    //Initialize SDL
    if( !SDL_Init( SDL_INIT_VIDEO ) )
    {
        SDL_Log( "SDL could not initialize! SDL error: %s\n", SDL_GetError() );
        success = false;
    }
    else
    {
        //Create window with renderer
        if( !SDL_CreateWindowAndRenderer( "SDL3 Tutorial: Color Modulation and Blending", kScreenWidth, kScreenHeight, 0, &gWindow, &gRenderer ) )
        {
            SDL_Log( "Window could not be created! SDL error: %s\n", SDL_GetError() );
            success = false;
        }
        else
        {
            //Initialize PNG loading
            int imgFlags = IMG_INIT_PNG;
            if( !( IMG_Init( imgFlags ) & imgFlags ) )
            {
                SDL_Log( "SDL_image could not initialize! SDL_image error: %s\n", SDL_GetError() );
                success = false;
            }
        }
    }

    return success;
}

bool loadMedia()
{
    //File loading flag
    bool success{ true };

    //Load scene images
    if( success &= gColorsTexture.loadFromFile( "07-color-modulation-and-alpha-blending/colors.png" ); !success )
    {
        SDL_Log( "Unable to load colors image!\n");
    }

    return success;
}

void close()
{
    //Clean up textures
    gColorsTexture.destroy();
    
    //Destroy window
    SDL_DestroyRenderer( gRenderer );
    gRenderer = nullptr;
    SDL_DestroyWindow( gWindow );
    gWindow = nullptr;

    //Quit SDL subsystems
    IMG_Quit();
    SDL_Quit();
}

int main( int argc, char* args[] )
{
    //Final exit code
    int exitCode{ 0 };

    //Initialize
    if( !init() )
    {
        SDL_Log( "Unable to initialize program!\n" );
        exitCode = 1;
    }
    else
    {
        //Load media
        if( !loadMedia() )
        {
            SDL_Log( "Unable to load media!\n" );
            exitCode = 2;
        }
        else
        {
            //The quit flag
            bool quit{ false };

            //The event data
            SDL_Event e;
            SDL_zero( e );
            
            //Set color constants
            constexpr int kColorMagnitudeCount = 3;
            constexpr Uint8 kColorMagnitudes[ kColorMagnitudeCount ] = { 0x00, 0x7F, 0xFF };
            enum eColorChannel
            {
                eColorChannelTextureRed       = 0,
                eColorChannelTextureGreen     = 1,
                eColorChannelTextureBlue      = 2,
                eColorChannelTextureAlpha     = 3,

                eColorChannelBackgroundRed    = 4,
                eColorChannelBackgroundGreen  = 5,
                eColorChannelBackgroundBlue   = 6,

                kColorChannelTotal            = 7,
                kColorChannelUnknown          = 8
            };

            //Initialize colors
            Uint8 colorChannelsIndices[ kColorChannelTotal ];
            colorChannelsIndices[ eColorChannelTextureRed ]   = 2;
            colorChannelsIndices[ eColorChannelTextureGreen ] = 2;
            colorChannelsIndices[ eColorChannelTextureBlue ]  = 2;
            colorChannelsIndices[ eColorChannelTextureAlpha ] = 2;
            
            colorChannelsIndices[ eColorChannelBackgroundRed ]   = 2;
            colorChannelsIndices[ eColorChannelBackgroundGreen ] = 2;
            colorChannelsIndices[ eColorChannelBackgroundBlue ]  = 2;

            //Initialize blending
            gColorsTexture.setBlending( SDL_BLENDMODE_BLEND );

            //The main loop
            while( quit == false )
            {
                //Get event data
                while( SDL_PollEvent( &e ) )
                {
                    //If event is quit type
                    if( e.type == SDL_EVENT_QUIT )
                    {
                        //End the main loop
                        quit = true;
                    }
                    //On key press
                    else if( e.type == SDL_EVENT_KEY_DOWN )
                    {
                        //Check for channel key
                        eColorChannel channelToUpdate = kColorChannelUnknown;
                        switch( e.key.key )
                        {
                            //Update texture color 
                            case SDLK_A:
                            channelToUpdate = eColorChannelTextureRed;
                            break;
                            case SDLK_S:
                            channelToUpdate = eColorChannelTextureGreen;
                            break;
                            case SDLK_D:
                            channelToUpdate = eColorChannelTextureBlue;
                            break;
                            case SDLK_F:
                            channelToUpdate = eColorChannelTextureAlpha;
                            break;
                            
                            //Update background color
                            case SDLK_Q:
                            channelToUpdate = eColorChannelBackgroundRed;
                            break;
                            case SDLK_W:
                            channelToUpdate = eColorChannelBackgroundGreen;
                            break;
                            case SDLK_E:
                            channelToUpdate = eColorChannelBackgroundBlue;
                            break;
                        }

                        //If channel key was pressed
                        if( channelToUpdate != kColorChannelUnknown )
                        {
                            //Cycle through channel values
                            colorChannelsIndices[ channelToUpdate ]++;
                            if( colorChannelsIndices[ channelToUpdate ] >= kColorMagnitudeCount )
                            {
                                colorChannelsIndices[ channelToUpdate ] = 0;
                            }

                            //Write color values to console
                            SDL_Log("Texture - R:%d G:%d B:%d A:%d | Background - R:%d G:%d B:%d",
                                kColorMagnitudes[ colorChannelsIndices[ eColorChannelTextureRed ] ],
                                kColorMagnitudes[ colorChannelsIndices[ eColorChannelTextureGreen ] ],
                                kColorMagnitudes[ colorChannelsIndices[ eColorChannelTextureBlue ] ],
                                kColorMagnitudes[ colorChannelsIndices[ eColorChannelTextureAlpha ] ],
                                kColorMagnitudes[ colorChannelsIndices[ eColorChannelBackgroundRed ] ],
                                kColorMagnitudes[ colorChannelsIndices[ eColorChannelBackgroundGreen ] ],
                                kColorMagnitudes[ colorChannelsIndices[ eColorChannelBackgroundBlue ] ]
                            );
                        }
                    }
                }

                //Fill the background
                SDL_SetRenderDrawColor( gRenderer,
                    kColorMagnitudes[ colorChannelsIndices[ eColorChannelBackgroundRed ] ],
                    kColorMagnitudes[ colorChannelsIndices[ eColorChannelBackgroundGreen ] ],
                    kColorMagnitudes[ colorChannelsIndices[ eColorChannelBackgroundBlue ] ],
                    0xFF );
                SDL_RenderClear( gRenderer );

                //Set texture color and render
                gColorsTexture.setColor(
                    kColorMagnitudes[ colorChannelsIndices[ eColorChannelTextureRed ] ],
                    kColorMagnitudes[ colorChannelsIndices[ eColorChannelTextureGreen ] ],
                    kColorMagnitudes[ colorChannelsIndices[ eColorChannelTextureBlue ] ]
                );
                gColorsTexture.setAlpha( kColorMagnitudes[ colorChannelsIndices[ eColorChannelTextureAlpha ] ]);
                gColorsTexture.render( ( kScreenWidth - gColorsTexture.getWidth() ) / 2.f, ( kScreenHeight - gColorsTexture.getHeight() ) / 2.f );

                //Update screen
                SDL_RenderPresent( gRenderer );
            } 
        }
    }

    //Clean up
    close();

    return exitCode;
}
